import os
import string
import sys
from random import SystemRandom

import flask
import jinja2
from flask_response_builder import encoders
from werkzeug.middleware.lint import LintMiddleware
from werkzeug.middleware.profiler import ProfilerMiddleware

from .config import FLASK_APP
from .converters import CONVERTERS


class AppFactory:
    """Flask app factory"""

    """
    default app name
    """
    app_name = FLASK_APP

    """
    default secret key file name
    """
    secret_file = '.secret.key'

    """
    custom url converters
    """
    url_converters = CONVERTERS

    """
    custom app config module
    """
    conf_module = 'flaskel.config'

    """
    additional options
    """
    options = {
        "json_encoder": encoders.JsonEncoder
    }

    def __init__(self):
        """

        """
        self._app = None

    def generate_secret_key(self, secret_file, key_length):
        """

        :param secret_file:
        :param key_length:
        """
        secret_key = ''.join(
            SystemRandom().choice(string.printable) for _ in range(key_length)
        )

        with open(secret_file, 'w') as f:
            f.write(secret_key)
            abs_file = os.path.abspath(secret_file)
            mess = "new secret key generated: take care of this file:{}"
            self._app.logger.warning(mess.format(abs_file))
        return secret_key

    def load_secret_key(self, secret_file):
        """

        :param secret_file:
        :return:
        """
        if not os.path.isfile(secret_file):
            return None

        with open(secret_file, 'r') as f:
            secret_key = f.read()
            self._app.logger.info("load secret key from: {}".format(secret_file))
        return secret_key

    def set_secret_key(self):
        """

        """
        secret_key = None
        key_length = self._app.config['SECRET_KEY_MIN_LENGTH']

        if not self._app.config.get('FLASK_ENV', '').lower().startswith('dev'):
            secret_file = self._app.config.get('SECRET_KEY')

            if secret_file:
                secret_key = self.load_secret_key(secret_file) or secret_file
            else:
                secret_file = self.secret_file
                secret_key = self.load_secret_key(secret_file)
                secret_key = secret_key or self.generate_secret_key(secret_file, key_length)
        elif not self._app.config.get('SECRET_KEY'):
            self._app.logger.debug('set secret key in development mode')
            secret_key = 'fake_very_complex_string'

        self._app.config['SECRET_KEY'] = secret_key or self._app.config['SECRET_KEY']
        if len(secret_key) < key_length:
            self._app.logger.warning("secret key length is less than: {}".format(key_length))

    def register_extensions(self, extensions):
        """

        :param extensions: custom extension added to defaults
        """
        with self._app.app_context():
            for name, e in (extensions or {}).items():
                try:
                    ext = e[0]
                    opt = e[1] if len(e) > 1 else {}
                    if not ext:
                        raise TypeError("extension could not be None")
                except (TypeError, IndexError) as exc:
                    mess = "Invalid extension '{}' configuration '{}':\n{}"
                    self._app.logger.debug(mess.format(name, e, exc))
                    continue

                ext.init_app(self._app, **opt)
                mess = "Registered extension '{}' with options: {}"
                self._app.logger.debug(mess.format(name, str(opt)))

    def register_blueprints(self, blueprints):
        """

        :param blueprints: list of application's blueprints
        """
        for b in (blueprints or []):
            try:
                bp = b[0]
                opt = b[1] if len(b) > 1 else {}
                if not bp:
                    raise TypeError('blueprint could not be None')
            except (TypeError, IndexError) as exc:
                self._app.logger.debug("invalid blueprint configuration '{}':\n{}".format(b, exc))
                continue

            self._app.register_blueprint(bp, **opt)
            self._app.logger.debug("Registered blueprint '%s' with options: %s", bp.name, str(opt))

    def set_config(self, conf_module=None, conf_map=None):
        """

        :param conf_module:
        :param conf_map:
        :return:
        """
        self._app.config.from_object(conf_module or self.conf_module)
        self._app.config.from_mapping(**(conf_map or {}))
        self._app.config.from_envvar('APP_CONFIG_FILE', silent=True)

    def register_converters(self, converters):
        """

        :param converters:
        """
        conv = {**self.url_converters, **(converters or {})}
        self._app.url_map.converters.update(conv)

        for k in conv.keys():
            self._app.logger.debug("Registered converter: '{}'".format(k))

    def register_template_folders(self, folders=()):
        """

        :param folders:
        """
        loaders = [self._app.jinja_loader]

        for fsl in folders:
            loaders.append(jinja2.FileSystemLoader(fsl))

        self._app.jinja_loader = jinja2.ChoiceLoader(loaders)
        self._app.logger.debug("Registered template folders: '{}'".format(", ".join(folders)))

    def register_middlewares(self, middlewares):
        """

        :param middlewares:
        """
        for middleware in (middlewares or []):
            self._app.wsgi_app = middleware(self._app.wsgi_app)
            self._app.logger.debug("Registered middleware: '{}'".format(middleware))

    def patch_app(self):
        """

        :return:
        """
        encoder = self.options.get('json_encoders')
        if encoder:
            self._app.json_encoder = encoder

        if self._app.config.get('DEBUG'):
            if self._app.config['WSGI_WERKZEUG_LINT_ENABLED']:
                self._app.wsgi_app = LintMiddleware(self._app.wsgi_app)

            if self._app.config['WSGI_WERKZEUG_PROFILER_ENABLED']:
                stream = self._app.config['WSGI_WERKZEUG_PROFILER_FILE']
                if stream:
                    stream = open(self._app.config['WSGI_WERKZEUG_PROFILER_FILE'], 'w')
                else:
                    stream = sys.stdout
            else:
                stream = None

            self._app.wsgi_app = ProfilerMiddleware(
                self._app.wsgi_app,
                stream=stream,
                restrictions=self._app.config['WSGI_WERKZEUG_PROFILER_RESTRICTION']
            )

        @self._app.before_first_request
        def create_database():
            sqla = self._app.extensions.get('sqlalchemy')
            if sqla is not None:
                sqla.db.create_all()

    def create(self, conf_module=None, conf_map=None, extensions=None,
               converters=(), blueprints=(), folders=(), middlewares=(), **kwargs):
        """

        :param conf_module: python module file
        :param conf_map: mapping configuration
        :param converters: custom url converter mapping
        :param extensions: custom extension appended to defaults
        :param blueprints: list of application's blueprints
        :param middlewares: list of wsgi middleware
        :param folders: list of custom jinja2's template folders
        :param kwargs: passed to Flask class
        :return:
        """
        self._app = flask.Flask(self.app_name, **kwargs)

        self.set_config(conf_module, conf_map)
        self.set_secret_key()

        self.register_template_folders(folders)
        self.register_converters(converters)
        self.register_extensions(extensions)
        self.register_blueprints(blueprints)
        self.register_middlewares(middlewares)

        self.patch_app()

    def getOrCreate(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        if self._app is not None:
            return self._app

        self.create(**kwargs)

        return self._app



import os

import pytest
from flask.testing import FlaskClient

from flaskel import AppFactory
from flaskel.ext import BASE_EXTENSIONS
from flaskel.ext.crypto import Argon2
from flaskel.ext.healthcheck import health_checks, health_mongo, health_redis, health_sqlalchemy
from flaskel.ext.sqlalchemy import db as dbsqla
from flaskel.ext.useragent import UserAgent
from flaskel.patch import ForceHttps
from tests.blueprints import BLUEPRINTS

BASE_DIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
SKEL_DIR = os.path.join(BASE_DIR, 'skeleton')
CONF_DIR = os.path.join(SKEL_DIR, 'config')


class TestClient(FlaskClient):
    """
        Implement this to customize Flask client
        Form example:
                def fetch(self, url, *args, **kwargs):
                    return self.open(url, method='FETCH', *args, **kwargs)
    """


@pytest.fixture
def app_prod():
    """

    """
    os.environ['APP_CONFIG_FILE'] = os.path.join(CONF_DIR, 'production.py')

    def test_health_true(**kwargs):
        return True, None

    def test_health_false(**kwargs):
        return False, "error"

    extra_ext = {
        "empty": (None, (None,)),  # NB: needed to complete coverage
        "useragent": (UserAgent(),),
        "argon2": (Argon2(),),
        "health_checks": (health_checks, {'extensions': (
            {
                'name': 'health_true',
                'func': test_health_true,
            },
            {
                'func': test_health_false,
            },
        )},),
    }

    health_checks.register('mongo', db=dbsqla)(health_mongo)
    health_checks.register('redis', db=dbsqla)(health_redis)
    health_checks.register('sqlalchemy', db=dbsqla)(health_sqlalchemy)

    _app = AppFactory().getOrCreate(
        conf_map=dict(TESTING=True),
        blueprints=(*BLUEPRINTS, *(None,), *((None,),)),  # NB: needed to complete coverage
        extensions={**BASE_EXTENSIONS, **extra_ext},
        folders=["skeleton/templates"],
        static_folder="skeleton/static"
    )

    _app.wsgi_app = ForceHttps(_app.wsgi_app)
    _app.test_client_class = TestClient
    _app.config['USER_AGENT_AUTO_PARSE'] = True
    return _app


@pytest.fixture
def app_dev():
    """

    """
    os.environ['APP_CONFIG_FILE'] = os.path.join(CONF_DIR, 'development.py')

    _app = AppFactory().getOrCreate(
        blueprints=BLUEPRINTS,
        extensions=BASE_EXTENSIONS,
        template_folder="skeleton/templates",
        static_folder="skeleton/static"
    )

    _app.test_client_class = TestClient
    return _app


@pytest.fixture
def testapp(app_prod):
    """

    :param app_prod:
    :return:
    """
    return app_prod.test_client()



from flask.views import View
from flask import render_template


# noinspection PyMethodMayBeStatic
class RenderTemplate(View):
    methods = ['GET', 'POST']

    def __init__(self, template):
        """

        :param template:
        """
        self._template = template

    def service(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        return kwargs

    def dispatch_request(self):
        """

        :return:
        """
        params = self.service()
        return render_template(self._template, **params)

    @classmethod
    def register(cls, app, name, url, **kwargs):
        """

        :param app:
        :param name:
        :param url:
        :param kwargs:
        """
        view_func = cls.as_view(name, **kwargs)
        app.add_url_rule(url, view_func=view_func)



from datetime import datetime

from flaskel.config import DATE_ISO_FORMAT


def from_iso_format(str_date, fmt, exc=True):
    """

    :param str_date: input date with ISO format
    :param fmt: format input date
    :param exc: raise or not exception (default True)
    :return: return formatted date
    """
    if not (exc and str_date):
        return None

    try:
        date = datetime.strptime(str_date, DATE_ISO_FORMAT)
        return date.strftime(fmt)
    except (ValueError, TypeError):
        if exc is True:
            raise


def to_iso_format(str_date, fmt, exc=True):
    """

    :param str_date: input date
    :param fmt: format input date
    :param exc: raise or not exception (default True)
    :return: return date with ISO format
    """
    if not (exc and str_date):
        return None

    try:
        return datetime.strptime(
            str_date, fmt
        ).strftime(DATE_ISO_FORMAT)
    except (ValueError, TypeError):
        if exc is True:
            raise



# noinspection PyPep8Naming,PyMethodMayBeStatic
class MyJsonRPC:
    def testAction1(self):
        return {
            "action1": True
        }

    def testAction2(self):
        return {
            "action2": True
        }



from werkzeug.urls import url_decode
from werkzeug.middleware.proxy_fix import ProxyFix


class ForceHttps(object):
    """
    NOTE: use this only if you can not touch reverse proxy configuration
    use ReverseProxied instead
    """
    def __init__(self, app):
        """

        :param app:
        """
        self._app = app

    def __call__(self, environ, start_response):
        """

        :param environ:
        :param start_response:
        :return:
        """
        environ['wsgi.url_scheme'] = 'https'
        return self._app(environ, start_response)


class ReverseProxied(ProxyFix):
    """
    adjust the WSGI environ based on X-Forwarded- that proxies in
    front of the application may set:

    -   X-Forwarded-For    -> sets REMOTE_ADDR
    -   X-Forwarded-Proto  -> sets wsgi.url_scheme
    -   X-Forwarded-Host   -> sets HTTP_HOST, SERVER_NAME, and SERVER_PORT
    -   X-Forwarded-Port   -> sets HTTP_HOST and SERVER_PORT
    -   X-Forwarded-Prefix -> sets SCRIPT_NAME

    for example nginx:
        location /prefix {
            proxy_pass http://127.0.0.1:5000;
            proxy_set_header Host $host;
            proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
            proxy_set_header X-Forwarder-Proto $proto;
            proxy_set_header X-Forwarder-Port 5000;
            proxy_set_header X-Forwarder-Prefix /prefix;
        }

    You must tell the middleware how many proxies set each header so it
    knows what values to trust. It is a security issue to trust values
    that came from the client rather than a proxy.

    The original values of the headers are stored in the WSGI
    environ as werkzeug.proxy_fix.orig, a dict.
    """
    def __init__(self, app, x_for=1, x_proto=1, x_host=1, x_port=0, x_prefix=1):
        """

        :param app:
        :param x_for:
        :param x_proto:
        :param x_host:
        :param x_port:
        :param x_prefix:
        """
        super().__init__(app, x_for, x_proto, x_host, x_port, x_prefix)


class HTTPMethodOverride(object):
    """
    Implements the hidden HTTP method technique.
    Not all web browsers or reverse proxy supports every HTTP method.
    Client can use 'X-HTTP-Method-Override' header or '_method_override' in querystring.
    Only POST method can be overridden because with GET requests
    could result unexpected behavior due to caching.
    """
    def __init__(self, app):
        """

        :param app:
        """
        self.app = app

    def __call__(self, environ, start_response):
        """

        :param environ:
        :param start_response:
        :return:
        """
        method = environ.get('HTTP_X_HTTP_METHOD_OVERRIDE', None)

        if '_method_override' in environ.get('QUERY_STRING', ''):
            args = url_decode(environ['QUERY_STRING'])
            method = args.get('_method_override')

        if environ['REQUEST_METHOD'] == 'POST' and method:
            environ['REQUEST_METHOD'] = method.upper()

        return self.app(environ, start_response)



import socket

from flaskel import cap
from flask_mail import Message

from . import celery


@celery.task(bind=True, default_retry_delay=60)
def send_async_email(self, email_data):
    """

    :param self:
    :param email_data:
    """
    try:
        message = Message(
            email_data['subject'],
            sender=cap.config['MAIL_DEFAULT_SENDER'],
            recipients=[
                email_data['to']
            ]
        )
        message.body = email_data['body']
        mail = cap.extensions['mail']

        old_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(60)

        mail.send(message)
        socket.setdefaulttimeout(old_timeout)
    except OSError as exc:
        self.retry(exc=exc)



# noinspection PyUnresolvedReferences
from flask import current_app as cap

from .ext import BASE_EXTENSIONS
from .factory import AppFactory
from .standalone import Server
from .utils.http import http_status as httpcode
from .version import *
from .wsgi import WSGIBuiltin, WSGIFactory



from flaskel import Server

from blueprints import BLUEPRINTS
from ext import EXTENSIONS


if __name__ == '__main__':
    server = Server(blueprints=BLUEPRINTS, extensions=EXTENSIONS)
    server.run()



# import your blueprint here
#
from .api import api
from .web import web
from .spa import spa

BLUEPRINTS = (
    # (<Blueprint>, <dict>)
    (api,),
    (spa,),
    (web, {
        'url_prefix': '/'
    }),
)



import sys
import traceback

from .base import WSGIBuiltin

try:
    from .gunicorn import WSGIGunicorn
except ImportError as exc:
    WSGIGunicorn = None
try:
    from .gevent import WSGIGevent
except ImportError as exc:
    WSGIGevent = None
try:
    from .tornado import WSGITornado
except ImportError as exc:
    WSGITornado = None
try:
    from .twisted import WSGITwisted
except ImportError as exc:
    WSGITwisted = None
try:
    from .waitress import WSGIWaitress
except ImportError as exc:
    WSGIWaitress = None


class WSGIFactory:
    DEFAULT_WSGI = {
        'builtin':  WSGIBuiltin,
        'gunicorn': WSGIGunicorn,
        'gevent':   WSGIGevent,
        'tornado':  WSGITornado,
        'twisted':  WSGITwisted,
        'waitress': WSGIWaitress,
    }

    @classmethod
    def getInstance(cls, name, **kwargs):
        """

        :param name:
        :param kwargs:
        :return:
        """
        return cls.getClass(name)(**kwargs)

    @classmethod
    def getClass(cls, name):
        """

        :param name:
        :return:
        """
        if name not in cls.DEFAULT_WSGI:
            raise ValueError("unable to find wsgi server: '{}'".format(name))

        wsgi_class = cls.DEFAULT_WSGI.get(name)

        if wsgi_class is None:
            raise ImportError(traceback.TracebackException)
            # print(traceback.format_exc(), file=sys.stderr)
            # sys.exit(-1)

        return wsgi_class



from celery import Celery


class FlaskCelery:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        :param kwargs:
        """
        self.session = None

        if app is not None:
            self.init_app(app, **kwargs)

    def init_app(self, app, name, **kwargs):
        """

        :param app:
        :param name:
        :param kwargs:
        :return:
        """
        if not hasattr(app, 'extensions'):
            app.extensions = {}
        app.extensions['celery'] = self

        self.session = Celery(name, **kwargs)
        self.session.conf.update(app.config)

        class ContextTask(self.session.Task):
            abstract = True

            def __call__(self, *args, **kw):
                with app.app_context():
                    return self.celery.Task.__call__(self, *args, **kw)

        # noinspection PyPropertyAccess
        self.session.Task = ContextTask


celery = FlaskCelery()



import os
import sys

import click
import yaml
from yaml.error import YAMLError

from flaskel import AppFactory, BASE_EXTENSIONS
from flaskel.utils.misc import parse_value
from flaskel.wsgi import BaseApplication, WSGIFactory

wsgi_types = click.Choice(WSGIFactory.DEFAULT_WSGI, case_sensitive=False)


def option_as_dict(ctx, param, value):
    """

    :param ctx:
    :param param:
    :param value:
    :return:
    """
    ret = {}
    for opt in value:
        k, v = opt.split("=", 2)
        ret.update({k: parse_value(v)})
    return ret


class Server:
    opt_config_attr = dict(
        default=None,
        help='app yaml configuration file'
    )
    opt_log_config_attr = dict(
        default=None,
        help='alternative log yaml configuration file'
    )
    opt_bing_attr = dict(
        default='127.0.0.1:5000',
        help='address to bind',
        show_default=True
    )
    opt_debug_attr = dict(
        is_flag=True,
        flag_value=True,
        default=False,
        help='enable debug mode'
    )
    opt_wsgi_server_attr = dict(
        default=None,
        type=wsgi_types,
        help='name of wsgi server to use'
    )
    option_wsgi_config_attr = dict(
        default={},
        multiple=True,
        callback=option_as_dict,
        metavar="KEY=VAL",
        help='wsgi configuration'
    )

    def __init__(self, app=None, blueprints=None, extensions=None, **kwargs):
        """

        :param app:
        :param blueprints:
        :param extensions:
        :param kwargs:
        :return:
        """
        self._app = app or AppFactory().getOrCreate(
            blueprints=blueprints,
            extensions={**BASE_EXTENSIONS, **extensions},
            **kwargs
        )

    def _register_options(self, func):
        """

        :param func:
        :return:
        """

        @click.command()
        @click.option('-b', '--bind', **self.opt_bing_attr)
        @click.option('-d', '--debug', **self.opt_debug_attr)
        @click.option('-c', '--config', **self.opt_config_attr)
        @click.option('-l', '--log-config', **self.opt_log_config_attr)
        @click.option('-w', '--wsgi-server', **self.opt_wsgi_server_attr)
        @click.option('-D', '--wsgi-config', **self.option_wsgi_config_attr)
        def _options(*args, **kwargs):
            return func(*args, **kwargs)

        return _options

    def run(self):
        """

        """
        @self._register_options
        def _forever(config, log_config, bind, debug, wsgi_server, wsgi_config):
            self._serve_forever(
                app=self._app,
                bind=bind,
                debug=debug,
                config=config,
                log_config=log_config,
                wsgi_server=wsgi_server,
                wsgi_config=wsgi_config
            )

        _forever()

    @staticmethod
    def _prepare_config(filename, **kwargs):
        """

        :param filename:
        :return: config
        """
        if filename is not None:
            try:
                with open(filename) as f:
                    config = yaml.safe_load(f)
            except (OSError, YAMLError) as e:
                print(e, file=sys.stderr)
                sys.exit(os.EX_OSFILE)
        else:
            env = os.environ.get('FLASK_ENV')
            env = env or ('development' if kwargs['debug'] else 'production')
            config = dict(
                app={'DEBUG': kwargs['debug'], 'ENV': env},
                wsgi={'bind': kwargs['bind'], 'debug': kwargs['debug']}
            )

        if kwargs['log_config'] is not None:
            config['app']['LOG_FILE_CONF'] = log_config

        # debug flag enabled overrides config file
        if kwargs['debug'] is True:
            config['app']['DEBUG'] = True

        return config

    def _serve_forever(self, config=None, log_config=None, bind=None, debug=None,
                       wsgi_class=None, wsgi_server=None, wsgi_config=None, **kwargs):
        """

        :param wsgi_class: optional a custom subclass of BaseApplication
        :param config: app and wsgi configuration file
        :param log_config: log configuration file
        :param bind: address to bind
        :param debug: enable debug mode
        :param wsgi_server: wsgi server chose
        :param wsgi_config: wsgi configuration
        :param kwargs: parameters passed to factory
        :return: never returns
        """
        config = self._prepare_config(config, debug=debug, bind=bind, log_config=log_config)

        if wsgi_server:
            wsgi_class = WSGIFactory.getClass(wsgi_server)
        elif wsgi_class and not issubclass(wsgi_class, BaseApplication):
            mess = '{} must be subclass of {}'
            raise TypeError(mess.format(wsgi_class, BaseApplication))
        else:
            wsgi_class = WSGIFactory.getClass('builtin')

        config.update(dict(wsgi=wsgi_config))
        kwargs['conf_map'] = config.get('app', {})
        wsgi = wsgi_class(self._app, options=config.get('wsgi', {}))

        wsgi.run()  # run forever



from flask_sqlalchemy import Model, SQLAlchemy


class SQLAModel(Model):
    def get_one(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        res = self.query.filters(**kwargs).get_or_404()
        return res.to_dict()

    def get_list(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        res_list = []
        for r in self.query.filters(**kwargs).all():
            res_list.append(r.to_dict())
        return res_list

    def to_dict(self, restricted=False):
        """

        :param restricted:
        """
        raise NotImplemented


db = SQLAlchemy(model_class=SQLAModel)



import os

import werkzeug.exceptions

from flaskel import httpcode
from flaskel.utils import http, uuid
from flaskel.utils.http.rpc import RPCInvalidRequest, RPCMethodNotFound, RPCParseError
# noinspection PyUnresolvedReferences
from tests import app_dev, app_prod, testapp


def test_app_dev(app_dev):
    client = app_dev.test_client()
    assert app_dev.config['FLASK_ENV'] == 'development'
    assert app_dev.config['SECRET_KEY'] == 'fake_very_complex_string'

    res = client.get('/')
    assert res.status_code == httpcode.SUCCESS

    res = client.get('/test_https', base_url='http://' + app_dev.config['SERVER_NAME'])
    assert res.status_code == httpcode.SUCCESS
    data = res.get_json()

    assert data['scheme'] == 'http'
    assert data['url_for'].startswith('http')


def test_app_runs(testapp):
    res = testapp.get('/')
    assert res.status_code == httpcode.SUCCESS


def test_app_return_html(testapp):
    res = testapp.get('/web')
    assert 'text/html' in res.headers['Content-Type']


def test_app_returns_json(testapp):
    res = testapp.get('/', base_url='http://api.' + testapp.application.config['SERVER_NAME'])
    assert res.status_code == httpcode.NOT_FOUND
    assert 'application/problem+json' in res.headers['Content-Type']


def test_api_resources(testapp):
    base_url = 'http://api.' + testapp.application.config['SERVER_NAME']

    res = testapp.get('/resources', base_url=base_url, headers={'Accept': 'application/xml'})
    assert res.status_code == httpcode.SUCCESS
    assert 'application/xml' in res.headers['Content-Type']

    res = testapp.get('/resources/1', base_url=base_url)
    assert res.status_code == httpcode.SUCCESS
    assert 'application/json' in res.headers['Content-Type']

    res = testapp.get('/resources/1/items', base_url=base_url)
    assert res.status_code == httpcode.SUCCESS
    assert 'application/json' in res.headers['Content-Type']

    res = testapp.get('/resources/1/not-found', base_url=base_url)
    assert res.status_code == httpcode.NOT_FOUND
    assert 'application/problem+json' in res.headers['Content-Type']

    res = testapp.delete('/resources/1', base_url=base_url)
    assert res.status_code == httpcode.NO_CONTENT

    data = {'item': 'test'}
    res = testapp.put('/resources/1', json=data, base_url=base_url)
    assert res.status_code == httpcode.SUCCESS

    res = testapp.post('/resources', json=data, base_url=base_url)
    assert res.status_code == httpcode.CREATED

    res = testapp.post('/resources/1/items', json=data, base_url=base_url)
    assert res.status_code == httpcode.CREATED


def test_api_cors(testapp):
    res = testapp.get('/resources', base_url='http://api.' + testapp.application.config['SERVER_NAME'])
    assert res.headers['Access-Control-Allow-Origin'] == '*'


def test_dispatch_error_web(testapp):
    res = testapp.get('/web-page-not-found')
    assert res.status_code == httpcode.NOT_FOUND
    assert 'text/html' in res.headers['Content-Type']


def test_dispatch_error_api(testapp):
    res = testapp.get('/api-not-found', base_url='http://api.' + testapp.application.config['SERVER_NAME'])
    assert res.status_code == httpcode.NOT_FOUND
    assert res.headers['Content-Type'] == 'application/problem+json'


def test_force_https(testapp):
    res = testapp.get('/test_https', base_url='http://' + testapp.application.config['SERVER_NAME'])
    assert res.status_code == httpcode.SUCCESS
    data = res.get_json()

    assert data['scheme'] == 'https'
    assert data['url_for'].startswith('https')


def test_reverse_proxy(testapp):
    res = testapp.get('/proxy', headers={
        'X-Forwarded-Prefix': '/test'
    })
    data = res.get_json()

    assert res.status_code == httpcode.SUCCESS
    assert data['script_name'] == '/test'
    assert data['original']['SCRIPT_NAME'] == ''


def test_secret_key_prod(testapp):
    assert testapp.application.config['FLASK_ENV'] == 'production'
    assert os.path.isfile('.secret.key')
    os.remove('.secret.key')


def test_method_override(testapp):
    res_header = testapp.post(
        '/method_override',
        headers={'X-HTTP-Method-Override': 'PUT'}
    )
    assert res_header.status_code == httpcode.SUCCESS

    res_query_string = testapp.post(
        '/method_override?_method_override=PUT'
    )
    assert res_query_string.status_code == httpcode.SUCCESS


def test_converters(testapp):
    res = testapp.get('/list/a-b-c')
    assert res.status_code == httpcode.SUCCESS
    assert len(res.get_json()) == 3


def test_utils_get_json(testapp):
    res = testapp.post('/invalid-json')
    assert res.status_code == httpcode.BAD_REQUEST


def test_utils_send_file(testapp):
    res = testapp.get('/download?filename=MANIFEST.in')

    assert res.status_code == httpcode.SUCCESS
    assert res.headers.get('Content-Disposition') == 'attachment; filename=MANIFEST.in'
    assert res.headers.get('X-Sendfile').endswith('./MANIFEST.in')
    assert res.headers.get('X-Accel-Redirect') == './MANIFEST.in'
    assert res.data == b''

    res = testapp.get('/download?filename=nofile.txt')
    assert res.status_code == httpcode.NOT_FOUND


def test_utils_uuid(testapp):
    res = testapp.get('/uuid')
    data = res.get_json()
    assert uuid.check_uuid('fake uuid') is False
    assert uuid.check_uuid(data.get('uuid1'), ver=1) is True
    assert uuid.check_uuid(data.get('uuid3'), ver=3) is True
    assert uuid.check_uuid(data.get('uuid4')) is True
    assert uuid.check_uuid(data.get('uuid5'), ver=5) is True


def test_crypto(testapp):
    passwd = 'my-favourite-password'
    crypto = testapp.application.extensions['argon2']

    res = testapp.get('/crypt/{}'.format(passwd))
    assert crypto.verify_hash(res.data, passwd)
    assert crypto.verify_hash(res.data, "wrong-pass") is False


def test_utils_http_client_simple(testapp):
    api = http.HTTPClient("http://httpbin.org", token='pippo', logger=testapp.application.logger)

    with testapp.application.app_context():
        res = api.delete('/status/200')
        assert res['status'] == httpcode.SUCCESS
        res = api.patch('/status/400')
        assert res['status'] == httpcode.BAD_REQUEST


def test_utils_http_client_exception(testapp):
    api = http.HTTPClient("http://httpbin.org", token='pippo', raise_on_exc=True)
    fake_api = http.HTTPClient('localhost', username='test', password='test')

    with testapp.application.app_context():
        res = fake_api.put('/', timeout=0.1)
        assert res['status'] == httpcode.SERVICE_UNAVAILABLE

        try:
            api.request('/status/500', 'PUT')
        except http.client.http_exc.HTTPError as exc:
            assert exc.response.status_code == httpcode.INTERNAL_SERVER_ERROR

        try:
            fake_api.request('/', timeout=0.1)
        except werkzeug.exceptions.HTTPException as exc:
            assert exc.code == httpcode.INTERNAL_SERVER_ERROR


def test_utils_http_client_filename(testapp):
    from flaskel.utils import http
    api = http.HTTPClient("http://httpbin.org", token='pippo')

    with testapp.application.app_context():
        filename = "pippo.txt"
        res = api.get('/response-headers', params={
            'Content-Disposition': "attachment; filename={}".format(filename)
        })
        assert http.misc.get_response_filename(res['headers']) == filename

        res = api.post('/response-headers', params={
            'Content-Disposition': "filename={}".format(filename)
        })
        assert http.misc.get_response_filename(res['headers']) == filename

        res = api.post('/response-headers', params={
            'Content-Disposition': filename
        })
        assert http.misc.get_response_filename(res['headers']) is None


def test_utils_http_jsonrpc_client(testapp):
    from flask import json

    params = dict(a=1, b=2)

    with testapp.application.app_context():
        api = http.JsonRPCClient("http://httpbin.org", "/anything")
        res = api.request('method.test', params=params)

    data = json.loads(res['data'])
    assert data['jsonrpc'] == '2.0'
    assert data['id'] == api.request_id
    assert data['params'] == params


def test_healthcheck(testapp):
    res = testapp.get('/healthcheck')
    data = res.get_json()

    assert res.status_code == httpcode.SERVICE_UNAVAILABLE
    assert res.headers['Content-Type'] == 'application/health+json'
    assert data['status'] == 'fail'
    assert data['links'] == {'about': None}
    assert data['checks']['health_true']['status'] == 'pass'
    assert data['checks']['test_health_false']['status'] == 'fail'
    assert data['checks']['sqlalchemy']['status'] == 'fail'
    assert data['checks']['mongo']['status'] == 'fail'
    assert data['checks']['redis']['status'] == 'fail'


def test_api_jsonrpc_ok(testapp):
    call_id = 1
    base_url = 'http://api.' + testapp.application.config['SERVER_NAME']

    res = testapp.post('/rpc', base_url=base_url, json={
        "jsonrpc": "2.0",
        "method":  "MyJsonRPC.testAction1",
        "params":  None,
        "id":      call_id
    })

    data = res.get_json()
    assert res.status_code == 200
    assert data['jsonrpc'] == "2.0"
    assert data['id'] == call_id
    assert data['result']['action1'] is True

    res = testapp.post('/rpc', base_url=base_url, json={
        "jsonrpc": "2.0",
        "method":  "MyJsonRPC.testAction2"
    })
    assert res.status_code == 204


def test_api_jsonrpc_error(testapp):
    call_id = 1
    base_url = 'http://api.' + testapp.application.config['SERVER_NAME']

    res = testapp.post('/rpc', base_url=base_url, json={
        "jsonrpc": "2.0",
        "method":  "NotFoundMethod",
        "id":      call_id
    })

    data = res.get_json()
    assert res.status_code == 200
    assert data['error']['code'] == RPCMethodNotFound().code
    assert data['jsonrpc'] == "2.0"
    assert data['id'] == call_id
    assert len(data['error']['message']) > 0

    res = testapp.post('/rpc', base_url=base_url, json={})
    data = res.get_json()
    assert res.status_code == 200
    assert data['error']['code'] == RPCParseError().code

    res = testapp.post('/rpc', base_url=base_url, json={
        "jsonrpc": "2.0"
    })
    data = res.get_json()
    assert res.status_code == 200
    assert data['error']['code'] == RPCInvalidRequest().code


def test_useragent(testapp):
    res = testapp.get('/useragent')
    data = res.get_json()

    assert res.status_code == httpcode.SUCCESS
    all(i in data.keys() for i in (
        'browser', 'device', 'os', 'raw'
    ))



import os

DEBUG = False
TESTING = False
FLASK_ENV = 'production'
APP_HOST = "flask.local"
APP_PORT = 5000

SERVER_NAME = os.environ.get('SERVER_NAME') or "{}:{}".format(APP_HOST, APP_PORT)

# in production env secret key should be saved in a file
SECRET_KEY = os.environ.get('SECRET_KEY_FILE')

# overrides here because skeleton in the source code has a different path than the installation path
BASE_DIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
LOG_FILE_CONF = os.path.join(os.path.join(BASE_DIR, 'config'), 'log.yaml')



import functools
import inspect

from flask.views import View

import flaskel.utils.http.rpc as rpc
from flaskel import httpcode, cap
from flaskel.ext import builder
from flaskel.utils.http import get_json


class JSONRPCView(View):
    version = '2.0'
    separator = '.'
    methods = ['POST']
    decorators = [builder.response('json')]

    def __init__(self, operations=None):
        """

        :param operations: a dict of jsonrpc operations like:
            {
                None: {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                "method1": {
                    "action1": funct1,
                    "action2": funct2,
                    ...
                },
                ...
            }
        """
        self.operations = operations or {}

    def _validate_payload(self):
        """

        :return:
        """
        payload = get_json(True)
        if not payload:
            raise rpc.RPCParseError() from None

        if 'jsonrpc' not in payload or \
                'method' not in payload:
            raise rpc.RPCInvalidRequest() from None

        if payload['jsonrpc'] != self.version:
            mess = "jsonrpc version is {}".format(self.version)
            raise rpc.RPCInvalidRequest(message=mess) from None

        return payload

    def _get_action(self, method):
        """

        :param method:
        :return:
        """
        m = method.split(self.separator)
        if len(m) > 1:
            op, action = m[0], self.separator.join(m[1:])
        else:
            op, action = None, m[0]

        try:
            return self.operations[op][action]
        except (IndexError, TypeError, KeyError) as exc:
            raise rpc.RPCMethodNotFound()

    def dispatch_request(self):
        """

        :return:
        """
        response = dict(jsonrpc=self.version, id=None)

        try:
            payload = self._validate_payload()
            response['id'] = payload.get('id')

            action = self._get_action(payload['method'])
            response['result'] = action(**(payload.get('params') or {}))

            if 'id' not in payload:
                return None, httpcode.NO_CONTENT

        except rpc.RPCError as ex:
            response['error'] = ex.as_dict()
        except Exception as ex:
            cap.logger.exception(ex)
            mess = str(ex) if cap.config['DEBUG'] is True else None
            response['error'] = rpc.RPCInternalError(message=mess).as_dict()

        return response

    def load_from_object(self, obj):
        """

        :param obj:
        """
        for m in inspect.getmembers(obj, predicate=inspect.isroutine):
            if not m[0].startswith('_'):
                self.method(obj.__class__.__name__, m[0])(m[1])

    def method(self, name=None, operation=None):
        """

        :param name:
        :param operation:
        :return:
        """

        def _method(func):
            """

            :param func:
            :return:
            """
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                obj = {operation: func}
                if name not in self.operations:
                    self.operations[name] = obj
                else:
                    self.operations[name].update(obj)

            return wrapped()

        return _method

    def register(self, app, name, url, **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name:
        :param url:
        """
        view_func = self.__class__.as_view(name, self.operations, **kwargs)

        url = url.rstrip('/')
        if not url.startswith('/'):
            url = '/{}'.format(url)

        app.add_url_rule(url, view_func=view_func, methods=self.methods)



import flask
from flask.views import MethodView
from sqlalchemy.exc import IntegrityError

from flaskel import httpcode
from flaskel.ext import builder


class Resource(MethodView):
    @builder.on_accept()
    def get(self, res_id=None, sub_resource=None):
        """

        :param res_id:
        :param sub_resource:
        :return:
        """
        if res_id is None:
            return self.on_collection()

        if sub_resource is None:
            return self.on_get(res_id)

        try:
            _sub_resource = getattr(self, "sub_{}".format(sub_resource))
            return _sub_resource(res_id)
        except AttributeError:
            flask.abort(httpcode.NOT_FOUND)

    @builder.on_accept()
    def post(self, res_id=None, sub_resource=None):
        """

        :return:
        """
        if res_id is None:
            return self.on_post()
        try:
            _sub_resource_post = getattr(self, "sub_{}_post".format(sub_resource))
            return _sub_resource_post(res_id)
        except AttributeError:
            flask.abort(httpcode.NOT_FOUND)

    @builder.no_content
    def delete(self, res_id):
        """

        :param res_id:
        :return:
        """
        return self.on_delete(res_id)

    @builder.on_accept()
    def put(self, res_id):
        """

        :param res_id:
        :return:
        """
        return self.on_put(res_id)

    def on_get(self, res_id):
        """

        :param res_id:
        :return:
        """
        self._not_implemented()

    def on_post(self):
        """

        :return:
        """
        self._not_implemented()

    def on_put(self, res_id):
        """

        :param res_id:
        :return:
        """
        self._not_implemented()

    def on_delete(self, res_id):
        """

        :param res_id:
        :return:
        """
        self._not_implemented()

    def on_collection(self):
        """

        :return:
        """
        self._not_implemented()

    def _not_implemented(self):
        """

        """
        raise NotImplemented

    @classmethod
    def register(cls, app, name, url, pk_type='int', **kwargs):
        """

        :param app: Flask or Blueprint instance
        :param name:
        :param url:
        :param pk_type: type of res_id
        """
        view_func = cls.as_view(name, **kwargs)

        url = url.rstrip('/')
        if not url.startswith('/'):
            url = '/{}'.format(url)

        app.add_url_rule(url, view_func=view_func, methods=['GET', 'POST'])

        app.add_url_rule(
            '{url}/<{type}:res_id>'.format(url=url, type=pk_type),
            view_func=view_func,
            methods=['GET', 'PUT', 'DELETE']
        )

        app.add_url_rule(
            '{url}/<{type}:res_id>/<sub_resource>'.format(url=url, type=pk_type),
            view_func=view_func,
            methods=['GET', 'POST']
        )


class SQLAResource(Resource):
    def __init__(self, db, model):
        """

        :param db:
        :param model:
        """
        self._db = db
        self._model = model

    def on_get(self, res_id):
        """

        :param res_id:
        """
        res = self.query.get_or_404(res_id)
        return res.to_dict()

    def on_collection(self, **kwargs):
        """

        :return:
        """
        res_list = []
        for r in self._model.query.filters(**kwargs).all():
            res_list.append(r.to_dict(restricted=True))
        return res_list

    def on_post(self, payload=None):
        """

        :param payload:
        :return:
        """
        res = self._model(**(payload or {}))
        try:
            self._db.session.add()
            self._db.session.commit()
        except IntegrityError:
            self._db.session.rollback()
            flask.abort(httpcode.CONFLICT)
        finally:
            self._db.session.close()

        return res.to_dict(), httpcode.CREATED

    def on_delete(self, res_id):
        """

        :param res_id:
        """
        res = self._model.query.get_or_404(res_id)
        res.delete()
        self._db.session.commit()

    def on_put(self, res_id, payload=None):
        """

        :param res_id:
        :param payload:
        :return:
        """
        res = self._model.query.get_or_404(res_id)
        res.update(**(payload or {}))
        self._db.session.add(res)
        self._db.session.commit()
        return res.to_dict()



class FakeLogger:
    def debug(self, *args):
        pass

    def info(self, *args):
        pass

    def warning(self, *args):
        pass

    def error(self, *args):
        pass

    def exception(self, *args):
        pass


# noinspection PyMethodMayBeStatic
class DummyLogger:
    def debug(self, *args):
        print(*args)

    def info(self, *args):
        print(*args)

    def warning(self, *args):
        print(*args)

    def error(self, *args):
        print(*args)

    def exception(self, *args):
        print(*args)



import os

DEBUG = False
TESTING = False
APP_NAME = os.environ.get('APP_NAME') or "flaskel"
APP_HOST = os.environ.get('APP_HOST') or '127.0.0.1'
APP_PORT = os.environ.get('APP_PORT') or 5000
FLASK_APP = os.environ.get('FLASK_APP') or "app:app"
FLASK_ENV = os.environ.get('FLASK_ENV') or "production"

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

CONF_DIR = os.path.join(BASE_DIR, 'skeleton', 'config')
LOG_FILE_CONF = os.path.join(CONF_DIR, 'log.yaml')

SECRET_KEY_MIN_LENGTH = 256

REQUEST_METHODS = {
    'WITHOUT_BODY': [
        'GET', 'DELETE'
    ],
    'WITH_BODY': [
        'POST', 'PUT', 'PATCH'
    ]
}

ALLOWED_METHODS = REQUEST_METHODS['WITH_BODY'] + REQUEST_METHODS['WITHOUT_BODY']

AC_ALLOW_ORIGIN = []

RB_DEFAULT_ACCEPTABLE_MIMETYPES = [
    "application/json",
    "application/xml"
]

ERROR_PAGE = 'core/error.html'
LOG_LOGGER_NAME = FLASK_ENV
LOG_APP_NAME = APP_NAME

CF_STRICT_ACCESS = False

USE_X_SENDFILE = True
ENABLE_ACCEL = True
ACCEL_BUFFERING = True
ACCEL_CHARSET = 'utf-8'
ACCEL_LIMIT_RATE = 'off'
SEND_FILE_MAX_AGE_DEFAULT = 86400

WSGI_WERKZEUG_LINT_ENABLED = True
WSGI_WERKZEUG_PROFILER_ENABLED = True
WSGI_WERKZEUG_PROFILER_RESTRICTION = (0.1,)
WSGI_WERKZEUG_PROFILER_FILE = 'profiler.txt'

DATE_ISO_FORMAT = "%Y-%m-%dT%H:%M:%S"



import os
import sys
import shutil
from pathlib import Path

import click

INIT_CONTENT = """
# generated via cli
from .version import *
"""


@click.group()
def cli():
    """

    """
    pass


@cli.command(name='init')
@click.argument('name')
def init(name):
    """

    :param name: application name
    """
    import flaskel

    package_path = flaskel.__path__[0]
    source = os.path.join(package_path, 'skeleton')
    destination = name

    try:
        shutil.copytree(source, destination)

        init_file = Path(os.path.join(destination, '__init__.py'))
        init_file.write_text(INIT_CONTENT)

        if not os.path.isfile('setup.py'):
            shutil.move(os.path.join(destination, 'setup.py'), '.')

            setup_file = Path('setup.py')
            text = setup_file.read_text()
            text = text.replace('{skeleton}', name)
            setup_file.write_text(text)
        else:
            os.remove(os.path.join(destination, 'setup.py'))

        cli_file = Path(os.path.join(destination, 'cli.py'))
        text = cli_file.read_text()
        text = text.replace('from .ext', 'from {}.ext'.format(name))
        text = text.replace('from .blueprint', 'from {}.blueprint'.format(name))
        cli_file.write_text(text)
    except OSError as e:
        print('Unable to create new app. Error: %s' % str(e), file=sys.stderr)


if __name__ == '__main__':
    cli()



class RPCError(Exception):
    def __init__(self, code, message, data=None):
        """

        :param code:
        :param message:
        :param data:
        """
        self.code = code
        self.message = message
        self.data = data

    def as_dict(self):
        """

        :return:
        """
        return dict(
            code=self.code,
            message=self.message,
            data=self.data
        )


class RPCParseError(RPCError):
    def __init__(self, message=None, data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(
            -32700,
            message or 'Invalid JSON was received by the server',
            data
        )


class RPCInvalidRequest(RPCError):
    def __init__(self, message=None, data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(
            -32600,
            message or 'The JSON sent is not a valid Request object',
            data
        )


class RPCMethodNotFound(RPCError):
    def __init__(self, message=None, data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(
            -32601,
            message or 'The method does not exist or is not available',
            data
        )


class RPCInvalidParams(RPCError):
    def __init__(self, message=None, data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(
            -32602,
            message or 'Invalid method parameter(s)',
            data
        )


class RPCInternalError(RPCError):
    def __init__(self, message=None, data=None):
        """

        :param message:
        :param data:
        """
        super().__init__(
            -32603,
            message or 'Internal JSON-RPC error',
            data
        )



import asyncio

import aiohttp

from . import misc
from .client import HTTPBase, httpcode


class HTTPBatchRequests(HTTPBase):
    def __init__(self, conn_timeout=60, read_timeout=60, **kwargs):
        """

        :param conn_timeout:
        :param read_timeout:
        :param kwargs:
        """
        super().__init__(**kwargs)
        self._conn_timeout = conn_timeout
        self._read_timeout = read_timeout

    async def http_request(self, session, **kwargs):
        """

        :param session:
        :param kwargs:
        :return:
        """
        try:
            async with session.request(**kwargs) as resp:
                # noinspection PyProtectedMember
                self._logger.info(misc.dump_request(resp._request_info))
                try:
                    body = await resp.json()
                except (aiohttp.ContentTypeError, ValueError, TypeError):
                    body = await resp.text()

                try:
                    resp.raise_for_status()
                except aiohttp.ClientResponseError as exc:
                    self._logger.warning(misc.dump_response(resp, self._dump_body))
                    if self._raise_on_exc is True:
                        raise

                    return dict(
                        body=body,
                        status=resp.status,
                        headers={k: v for k, v in resp.headers.items()}
                    )

                self._logger.info(misc.dump_response(resp, self._dump_body))
                return dict(
                    body=body,
                    status=resp.status,
                    headers={k: v for k, v in resp.headers.items()}
                )
        except aiohttp.ClientOSError as exc:
            self._logger.exception(exc)
            if self._raise_on_exc is True:
                raise

            return dict(
                body={},
                status=httpcode.SERVICE_UNAVAILABLE,
                headers={},
                exception=exc
            )

    async def _batch(self, requests):
        """

        :param requests:
        :return:
        """
        tasks = []
        async with aiohttp.ClientSession(
                conn_timeout=self._conn_timeout,
                read_timeout=self._read_timeout
        ) as session:
            for params in requests:
                tasks.append(self.http_request(session, **params))

            return await asyncio.gather(*tasks, return_exceptions=not self._raise_on_exc)

    def request(self, requests, **kwargs):
        """

        :param requests:
        :return:
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._batch(requests))



import os

DEBUG = True
TESTING = True

APP_HOST = "flask.local"
APP_PORT = 5000
FLASK_ENV = 'development'

SERVER_NAME = "{}:{}".format(APP_HOST, APP_PORT)

CF_STRICT_ACCESS = False

# overrides here because skeleton in the source code has a different path than the installation path
BASE_DIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
LOG_FILE_CONF = os.path.join(os.path.join(BASE_DIR, 'config'), 'log.yaml')



import flask

from . import http_status as httpcode


def get_json(allow_empty=False):
    """

    :param allow_empty:
    :return:
    """
    payload = flask.request.get_json()

    if not payload:
        if allow_empty:
            payload = {}
        else:
            flask.abort(httpcode.BAD_REQUEST, 'No JSON given')

    return payload


def send_file(directory, filename, **kwargs):
    """

    :param directory:
    :param filename:
    :param kwargs:
    """
    kwargs.setdefault('as_attachment', True)
    file_path = flask.safe_join(directory, filename)
    cap = flask.current_app

    try:
        resp = flask.send_file(file_path, **kwargs)
    except IOError as exc:
        cap.logger.warning(str(exc))
        flask.abort(httpcode.NOT_FOUND)
        return  # only to prevent warning

    # following headers works with nginx compatible proxy
    if cap.use_x_sendfile is True and cap.config.get('ENABLE_ACCEL') is True:
        resp.headers['X-Accel-Redirect'] = file_path
        resp.headers['X-Accel-Charset'] = cap.config['ACCEL_CHARSET']
        resp.headers['X-Accel-Limit-Rate'] = cap.config['ACCEL_LIMIT_RATE']
        resp.headers['X-Accel-Expires'] = cap.config['SEND_FILE_MAX_AGE_DEFAULT']
        resp.headers['X-Accel-Buffering'] = 'yes' if cap.config['ACCEL_BUFFERING'] else 'no'

    return resp


def dump_headers(hdr):
    """
    dumps http headers: useful for logging

    :param hdr: headers' dictionary
    :return: string representation of headers
            k1: v1
            k2: v2
    """
    return '\n'.join('{}: {}'.format(k, v) for k, v in hdr.items())


def dump_request(req, dump_body=None):
    """
    dump http request: useful for logging

    :param req: Request instance
    :param dump_body: flag to enable or disable dump of request's body (overrides default)
    :return: prettified representation of input as string
    """
    if dump_body is True:
        req_body = req.body
    else:
        req_body = "request body not dumped"

    return 'REQUEST: {} {}\nheaders:\n{}\nbody:\n{}'.format(
        req.method, req.url, dump_headers(req.headers), req_body
    )


def dump_response(response, dump_body=None):
    """
    dump http response: useful for logging

    :param response: Response instance
    :param dump_body: flag to enable or disable dump of response's body (overrides default)
    :return: prettified representation of input as string
    """
    resp = response
    hdr = resp.headers

    if dump_body is True:
        resp_body = get_response_filename(resp.headers) or resp.text
    else:
        resp_body = "response body not dumped"

    status_code = resp.status_code if hasattr(resp, 'status_code') else resp.status

    return 'RESPONSE status code: {}\nheaders:\n{}\nbody:\n{}'.format(
        status_code, dump_headers(hdr), resp_body
    )


def get_response_filename(headers):
    """
    get filename from Content-Disposition header.
    i.e.: attachment; filename="<file name.ext>"

    :param headers: http headers dict
    :return: only the file name
    """
    hdr = headers.get('Content-Disposition')
    if not hdr:
        return None

    tmp = hdr.split(';')
    hdr = tmp[1] if len(tmp) > 1 else tmp[0]
    tmp = hdr.split('=')

    if len(tmp) > 1:
        return tmp[1].strip('"').lstrip('<').rstrip('>')
    else:
        return None
